/**********************************************************************************/
/*                                                                                */ 
/* dcasvm_learn-0.03: Dual Coorinate Ascent SVM (ver. 0.03)                       */ 
/*                                                                                */ 
/* Author: Yutaka Sasaki, Toyota Technological Institute                          */ 
/*         (Yutaka.Sasaki<AT>toyota-ti.ac.jp)                                     */ 
/*                                                                                */ 
/* Licensed under the GNU General Public License, Version 2 (the "License")       */
/* You may not use this file except in compliance with the License.               */
/* You may obtain a copy of the License at                                        */
/*                                                                                */
/*      http://http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt               */
/*                                                                                */
/* Unless required by applicable law or agreed to in writing, software            */
/* distributed under the License is distributed on an "AS IS" BASIS,              */
/* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.       */
/* See the License for the specific language governing permissions and            */
/* limitations under the License.                                                 */
/**********************************************************************************/

/* Disclaimer: This is not a good C++ code. This is C implementation with C++     */
/* data type extension. Minimum Object-Oriented progamming is involved.           */

/* Header files */
#include <fstream>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctime>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include <map>
#include <set>
#include <vector>
#include <unistd.h>

#define MIN_ITERATION 300

char version[] = "Dual Coordinate Ascent L1-SVM (ver. 0.03) Y. Sasaki, TTI, Japan\n";
double* W;
std::vector< std::vector< std::pair< int, double > > >  X;
std::vector< int >  positive;
std::vector< int >  negative;
std::map< int, int > Y;
char* data_file;
char* model_file="model.dcasvm"; 
int n_iteration = 1000000;
double C=1.0;
int verbose=0;
int max_findex=2000;
int balanced=0;
double epsilon=0.1;
int block=100;

/* prototypes */
int dcal1_read_training_file ( char fname[] );
void dcal1_train_simple( int n_data, int iteration );
void dcal1_train( int n_data, int iteration );
void usage( char* command );
double dot_product( int docid );
double self_dot_product( int docid );
void write_W( char* fname );
double clip( double x, double L, double U );
void read_option (int argc, char *argv[]);
double gettimeofday_sec();

/* -------------------------------------------------------- */
/*                      Main program                        */
/* -------------------------------------------------------- */

int main(int argc, char *argv[]) {

  int i;
  char* command;
  int n_data;

  printf( version );

  /* command line options */
  command = argv[0];
  read_option(argc, argv);
  if ( data_file=='\0') { usage( command ); exit(1); }

  /* initialize */
  //  srand((unsigned int)time(NULL));

  /* print header */
  printf( "*** data file name = %s\n", data_file );
  printf( "*** model file name = %s\n", model_file );
  printf( "*** max iterations = %d\n", n_iteration );
  if (balanced) printf( "*** Sampling: balanced stochastic\n" );
  else          printf( "*** Sampling: stochastic\n" );
  printf( "*** epsilon = %g\n", epsilon );
  printf( "*** block size = %d\n", block );

  /* read data */
  n_data = dcal1_read_training_file ( data_file );
  printf( "*** #data = %d\n",n_data );

  /* train an SVM model */
  W = new double[max_findex+1];
  for( i=0; i<= max_findex ; i++ ) { W[i]=0; }
  dcal1_train( n_data, n_iteration );

  printf( "Training done.\n" );

  write_W( model_file );
  delete[] W;
}

/* -------------------------------------------------------- */
/*                      Subroutines                         */
/* -------------------------------------------------------- */

double gettimeofday_sec() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec + (double)tv.tv_usec*1e-6;
}

void dcal1_train( int N, int iteration ) {
  int rnd, d, k, t;
  int y;
  double UB = -1000;
  double LB =  1000;
  double* alpha;     /* Lagrange multiplier */
  double* Q;         /* Qii = x_i \dot x_i*/
  double G;          /* loss */ 
  double PG;
  double a;          /* old alpha[i] */
  int count=0;       /* loop counter */

  printf( "<< Training in\n" );
  printf( "%d positives and %d negaties\n", (int)positive.size(), (int)negative.size() );

  /* initialize */
  alpha = new double[N];
  Q = new double[N];

  double start_time = gettimeofday_sec();
  for ( k=0 ; k<N ; k++ )  {
     alpha[k] = 0.0; 
     Q[k] = self_dot_product(k);
  }

  /* outer loop */
  while( count < iteration ) {
     UB = -1000; LB = 1000;
     count++;
     /* inner loop */
     for ( t=0 ; t<block ; t++ ) {

        /* sampling */
        if (balanced) {
           if ( t % 2 == 0 ) { 
              rnd = rand() % positive.size(); // random number in 0..#pdata-1
              d = positive[rnd];  
              y=1;
           } else {
              rnd = rand() % negative.size();
              d = negative[rnd];
              y=-1;
           }
        } else {
           rnd = rand() % (positive.size()+negative.size());
           if ( rnd < (int)positive.size() ) {
              d = positive[rnd];  
              y=1;
           } else {
              d = negative[rnd-positive.size()];
              y=-1;
           }
        }

        /* loss and bound */
       G =  y*dot_product(d) - 1.0;
       PG = 0.0;
       if ( alpha[d] == 0.0 ) { 
          if ( G<0.0 ) PG = G; 
       }
       else if ( alpha[d] == C ) { 
          if ( G>0.0 ) PG = G; 
       } 
       else PG = G;

       if ( PG>UB ) UB=PG;
       if ( PG<LB ) LB=PG;

       /* update with L1 regularization */
       if ( PG != 0.0 ) {
         a = alpha[d];
         alpha[d] = clip( alpha[d] - G/Q[d], 0.0, C );

         /* revise W */
         std::vector< std::pair< int, double > >::iterator iter;
         for( iter = X[d].begin() ; iter != X[d].end() ; iter++ ) {
            W[iter->first] += (alpha[d] - a)*y*(iter->second);
         } /* end of for */
       } /* end of if */  
     } /* end of for */

     if (verbose) fprintf( stderr, "UB-LB=%g, epsilon=%g\n", UB-LB, epsilon );
     if ( (UB-LB < epsilon)&&(count > MIN_ITERATION) ) {
        printf( ">> Training out (%g sec) (%d iterations)\n", gettimeofday_sec()-start_time, count );  
        delete[] alpha;
        delete[] Q;
        return;
     }
  } /* end of while */
 printf( "Training out (%g sec) (%d iterations)\n", gettimeofday_sec()-start_time, count );  
}

int dcal1_read_training_file ( char fname[] ) {

  int iatt, icat=1;
  int line=0;
  register char *term;
  double fval;
  int docid = 0;
  register char *pc;
  std::vector< std::pair< int, double > >* x; 
  std::string strbuf;
  char* buf;

  printf( "Reading data file \"%s\" ... \n", fname);

  std::ifstream ifs;

  ifs.open( fname );
  if ( ifs.fail() ) { printf( "Not found!\n" ); exit(1); }

  /* process each line */
  while ( std::getline(ifs,strbuf) ) {
    line++;
    x = new std::vector< std::pair< int, double > >; 
    buf = new char[strbuf.size()+1];
    strcpy( buf, strbuf.c_str() );
    pc = buf;
    /* id 0 is  bias */
    x->push_back( std::pair<int,double>(0, 1.0) );

    /* read the class label */
    term = pc;
    pc = strchr( pc, ' ' );
    icat = atoi( term );

    Y[docid] = icat;
    if ( icat == 1 ) { positive.push_back(docid);}
    else { negative.push_back(docid); }

    /* read feature list */
    while ( *pc != '\0' ) {
        /* read feature id */
        pc++;
        if (*pc == '\n') break;
        iatt=0;
        while ( *pc != ':' ) { iatt = iatt*10+(*pc-48); pc++; }
        if ( iatt<0 ) fprintf( stderr,  "**** Error: nagatie value of xi=%d\n", iatt );
        if ( iatt>max_findex ) { max_findex = iatt; }
        /* read feature value */
        pc++;
        term = pc;
        pc = strchr( pc, ' ' );
        if ( pc == NULL ) {   
            fval = atof( term );
            x->push_back( std::pair<int,double>(iatt,fval) );
            break;
        } else {
            fval = atof( term );
            x->push_back( std::pair<int,double>(iatt,fval) );
	}
        if ( *(pc+1) == '\n') { break; }
    } /* end of while */

    delete[] buf;
    docid++;
    if ( docid % 10000 == 0 ) {  fprintf( stderr, "*** %d data read\n", docid); }

    /* store a sample to X */
    X.push_back( *x );

  } /* end of while */

  ifs.close();
  printf( "*** Largest feature ID = %d\n", max_findex );
  printf( "Done.\n" );
  return X.size();
}

void usage(char* command ) {
  fprintf( stderr, " %s\n", version );
  fprintf( stderr, " (Usage) %s -d data_file [options]\n",command );
  fprintf( stderr, "    Options:\n" );
  fprintf( stderr, "       -m <model file>\n" );
  fprintf( stderr, "       -d <training data file>\n" );
  fprintf( stderr, "       -i <#iterations> (max iterations, default 1,000,000)\n" );
  fprintf( stderr, "       -c <SVM hyper-parameter>\n" );
  fprintf( stderr, "       -e <epsion> (stopping criterion, default 0.5)\n" );
  fprintf( stderr, "       -b <block size> (mini batch size, default 100)\n" );
  fprintf( stderr, "       -v (verbose)\n" );
  fprintf( stderr, "       -s <sampling strategy number> (0:stochastic, 1:balanced, default 0)\n" );
}

double dot_product( int docid ) {
  register double r=0;
  std::vector< std::pair< int, double > >::iterator iter;

  for( iter = X[docid].begin() ; iter != X[docid].end() ; ++iter ) {
    r += W[iter->first]*(iter->second);
  }
  return r; 
}

double self_dot_product( int docid ) {
  register double r=0;
  if ( (int)X.size() <= docid ) return 0.0;

  std::vector< std::pair< int, double > >::iterator iter;
  for( iter = X[docid].begin() ; iter != X[docid].end() ; ++iter ) {
    r += (iter->second)*(iter->second);
  }
  return r; 
}

void write_W ( char *fname ) {
  FILE *fp = fopen( fname, "w" );

  printf( "Writing model file \"%s\" ... ", fname);
  fprintf( fp, "%.15g", W[0] );
  for( int j=1; j<= max_findex ; j++ ) { 
     if ( (W[j] > 0.0000000000001)||(W[j] < -0.0000000000001) ) { fprintf( fp, " %d:%.15g", j, W[j] ); }
  }
  fprintf( fp, "\n" );
  fclose( fp );
  printf( "Done.\n" );
}

double clip ( double x, double L, double U ) {
  if (x<L) return L;
  if (x>U) return U;
  return x;
}

void read_option (int argc, char *argv[]) {
  char option_char;

  /* options */
  /*   -m <model file> */
  /*   -d <training data file> */
  /*   -i <#iterations> */
  /*   -c <SVM hyper-parameter> */
  /*   -e <epsilon> */
  /*   -b <block size> */
  /*   -v (verbose) */
  /*   -s <sampling strategy> 0:stochastic, 1:balanced stochastic ) */

  while ((option_char = getopt(argc, argv, "m:d:i:c:b:e:s:v")) != EOF) {
    switch (option_char) {  
      case 'm': model_file = optarg; break;
      case 'd': data_file = optarg; break;
      case 'i': n_iteration  = atoi(optarg); break;
      case 'c': C = atof(optarg); 
                if (C>0) break; 
                else { fprintf( stderr, "C should be more than zero.\n"); exit(0); }
      case 'e': epsilon = atof(optarg); break;
      case 'b': block = atoi(optarg); break;
      case 'v': verbose = 1; break;
      case 's': balanced = atoi(optarg); break;
      case ':': usage( argv[0] ); exit(0);
      case '?': usage( argv[0] ); exit(0);
    }
  }
}

/**********************************************************************************/
